@@ -40,13 +40,17 @@ namespace snapper
     class SDir;
 
 
-    struct ConfigInfo : public SysconfigFile
+    class ConfigInfo : public SysconfigFile
     {
+    public:
+
 	explicit ConfigInfo(const string& config_name);
 
 	const string& getConfigName() const { return config_name; }
 	const string& getSubvolume() const { return subvolume; }
 
+	virtual void setValue(const string& key, const string& value);
+
     private:
 
 	const string config_name;
@@ -68,6 +72,12 @@ namespace snapper
 	virtual const char* what() const throw() { return "invalid config"; }
     };
 
+    struct InvalidConfigdataException : public SnapperException
+    {
+	explicit InvalidConfigdataException() throw() {}
+	virtual const char* what() const throw() { return "invalid configdata"; }
+    };
+
     struct InvalidUserdataException : public SnapperException
     {
 	explicit InvalidUserdataException() throw() {}
